# -*- coding: utf-8 -*-
"""
Created on Tue Aug 27 19:52:22 2019

@author: Reuben

This module includes the Container class upon which all of vartrix is built.

The idea of a container is to contain all of the key-value pairs for some
set of variables or parameters. Keys have an implied heirarchy using a
dot-format. For example, 'A.b' means that 'b' is a subkey of 'A'. All keys
are stored in a simple flat dictionary format, but some extra methods
allow the heirarchy to be used to simplify usage.

"""

import weakref
from collections import defaultdict
from contextlib import contextmanager

from . import utils

def is_root(dotkey):
    ''' Check if a dotkey is the root (container level) '''
    return dotkey in [None, '__ROOT__', '', '.']

def safe_root(dotkey):
    ''' Ensure root dotkeys are consistent '''
    if is_root(dotkey):
        return '__ROOT__'
    else:
        return dotkey


class Picker():
    def __init__(self, container, dotkey):
        self._container = container
        self._dotkey = dotkey
        
    def __getattr__(self, key):
        return self._container[self._dotkey + '.' + key]

    def __getitem__(self, key):
        return self._container[self._dotkey + '.' + key]


class Container(dict):
    ''' A dictionary-like class that updates observers 
    
    Args:
        dct (dict): A dictionary (possibly nested) of key-value pairs to use.
        name (str): The name of the container.
        
    Note:
        A `dotkey` is a dictionary key in the Container. It's called a dotkey
        simply because it's designed to have dots in it to represent a
        heirarchy.
    '''

    def __init__(self, dct=None, name=None):
        self.name = name
        self._observers = defaultdict(weakref.WeakSet)
        self._backup = {}
        if dct is not None:
            self.load(dct)

    def get(self, dotkey):
        ''' Return the value of a dotkey '''
        return self[dotkey]

    def lget(self, key_list):
        ''' Return the value of a dotkey specified by a list of strings
        
        Args:
            key_list (list([str])): The strings that define the dotkey. For
            example, `['a', 'b']` would equate to the dotkey `'a.b'`.
        '''
        return self['.'.join(key_list)]

    def dget(self, dotkeys):
        ''' Return a dictionary of dotkey-values for multiple dotkeys 
        
        Args:
            dotkeys (list[(str)]): A list of dotkey strings.
        '''
        dct = {}
        for dotkey in dotkeys:
            dct[dotkey] = self[dotkey]
        return dct
            
    def set(self, dotkey, val, safe=False):
        ''' Set the value of a dotkey 
        
        Args:
            dotkey (str): The key value
            val: The value to set. It can be a numpy array.
            safe (bool): Optional boolean. If true, the dotkey must already
            exist in the Container instance.
        '''
        if safe and dotkey not in self:
            raise KeyError('In safe mode, key ' + dotkey + ' must be present.')
        v = utils.denumpify(val)
        super().__setitem__(dotkey, v)
        self.update_observers(dotkey, v)
        
    def lset(self, key_list, val, safe=False):
        ''' Set the value of a dotkey specified as a list 
        
        Args:
            dotkey (list([str])): The list of strings for the dotkey.
            val: The value to set. It can be a numpy array.
            safe (bool): Optional boolean. If true, the dotkey must already
            exist in the Container instance.
        '''
        self.set('.'.join(key_list), val)

    def dset(self, dct, safe=False):
        ''' Set multiple values specified in a dictionary
        
        Args:
            dct (dict): The dictionary of dotkey-value pairs.
            safe (bool): Optional boolean. If true, the dotkey must already
            exist in the Container instance.
        '''
        for dotkey, val in dct.items():
            self.set(dotkey, val, safe=safe)

    def register_observer(self, dotkey, view):
        ''' Register an observer (typically a view) for a dotkey prefix
        
        Args:
            dotkey (str): The associated dotkey prefix
            view (View): A view instance (or other compatible object)
            
        Note:
            Here, the dotkey is for a higher level of the heirarchy. For
            example, if the container includes dotkeys of 'a.b.a', 'a.b.b', and
            'a.b.c', the dotkey specified here could be 'a.b'. It would 
            then ensure the view got data for all dotkeys beginning with 
            'a.b' (e.g. 'a.b.c'). Nested dotkeys are not included. 
            E.g. 'a.b' would not capture the dotkey of 'a.b.c.d'.
        '''
        self._observers[safe_root(dotkey)].add(view)

    def __missing__(self, key):
        try:
            ret = self.get_dct(key)
        except:
            return super().__missing__()
        return ret

    def get_dct(self, dotkey):
        ''' Return a dictionary of all key-value pairs for a dotkey prefix 
        
        Args:
            dotkey (str): The dotkey prefix. The dictionary will include all
            dotkeys that begin with this prefix.
            
        Note:
            Changes within nested dictionaries are not tracked.
        '''
        if is_root(dotkey):
            return self.copy()
        n = len(dotkey)
        out = {}
        for key, val in self.items():
            if key[:n] == dotkey and len(key) > n+1:
                # Check it's definately a dot after the dotkey before adding it
                # Otherwise it might add e.g. underscored dotkeys.
                if key[n] == '.':
                    k = key[n+1:]
                    out[k] = val
        return out

    def update_observers(self, dotkey, val):
        ''' Update the observers for a given dotkey with a value '''
        def all_possible(dotkey):
            ''' Include observes at higher levels '''
            split = dotkey.split('.')
            roots = ['__ROOT__']
            keys = [dotkey]
            for i in range(1, len(split)):
                roots.append('.'.join(split[:i]))
                keys.append('.'.join(split[i:]))
            return roots, keys
        for root, key in zip(*all_possible(dotkey)):
            for view in self._observers[root]:
                view._view_update(key, val)

    def __setitem__(self, key, val):
        return self.set(key, val)
    
    def _split_dotkey(self, dotkey):
        ''' Split a string dotkey into its root (prefix) and key '''
        split = dotkey.split('.')
        k = split.pop()
        root = safe_root('.'.join(split))
        return root, k

    def nested(self):
        ''' Create a nested dictionary representation of the container '''
        return utils.nested(self)
    
    def container(self, dct, parents=None, current=None):
        ''' Create a container dictionary (recursively) 
        
        Note:
            This method uses recursion to handle nested dictionaries.
        '''
        current = {} if current is None else current
        parents = [] if parents is None else parents
        for key, obj in dct.items():
            full_key = parents + [str(key)]
            if isinstance(obj, dict):
                self.container(obj, parents=full_key, current=current)
            else:
                current['.'.join(full_key)] = obj
        return current
    
    def load(self, dct):
        ''' Set the container data using a dictionary '''
        self.clear()
        self._backup.clear()
        self.add(dct)
        
    def reset(self):
        self.load(self._backup.copy())
        
    def add(self, dct):
        ''' Add another set of data to the container '''
        cdct = self.container(dct)
        self.update(cdct)
        self._backup.update(cdct)
        for dotkey, observers in self._observers.items():
            for observer in observers:
                observer.dirty = True
        for dotkey, observers in self._observers.items():
            for observer in observers:
                observer.check_refresh()
                
    @classmethod
    def combine(cls, dct):
        ''' Combine a dictionary of containers '''
        c = cls()
        for key, container in dct.items():
            for k, v in container.items():
                dotkey = key + '.' + k
                c[dotkey] = v
        return c
    
    @classmethod
    def merge(cls, containers):
        ''' Combine a dictionary of containers '''
        c = cls()
        for container in containers:
            c.update(container)
            for dotkey, observers in container._observers.items():
                c._observers[dotkey] |= observers # set union, in place
        return c
    
    @contextmanager
    def context(self, dct, safe=True):
        ''' A context manager for temporary changes in values 
        
        Args:
            dct (dict): A dictionary of dotkey-value pairs.
        '''
        originals = {k: self[k] for k in dct.keys()}
        self.dset(dct, safe=safe)
        yield self
        self.dset(originals, safe=safe)
        
    def get_picker(self, dotkey):
        return Picker(self, dotkey)
    
    def copy(self):
        return Container(dct=self, name=self.name)
    
    def to_dict(self):
        return dict(self)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 30 20:39:58 2019

@author: Reuben
"""

import unittest

from vartrix import utils


class Test_Utils(unittest.TestCase):
    def test_nest(self):
        ret = utils._nest(['a', 'b', 'c', 'd'], 5)
        expected = {'a': {'b': {'c': {'d': 5}}}}
        self.assertDictEqual(ret, expected)
        
    def test_nested(self):
        dct = {'a.b': 5, 'b.c': 7}
        ret = utils.nested(dct)
        expected = {'a': {'b': 5},
                    'b': {'c': 7}}
        self.assertDictEqual(ret, expected)


# -*- coding: utf-8 -*-
"""
Created on Tue Aug 27 20:59:39 2019

@author: Reuben

View instances make it easier to interact with containers. For example, 
if a container includes keys 'A.a' and 'A.b', we could make a view for all
keys beginning with 'A'. The view would then have keys 'a', and 'b', because
those are subkeys of 'A'.

"""

import uuid
from contextlib import contextmanager

from . import utils
from . import settings

def get_bases(obj):
    ''' Get a list of bases for an object instance '''
    def dotkey(c):
        long_name = c.__module__ + '.' + c.__name__
        split = long_name.split('.')
        return '.'.join(split[1:])
        
    bases = [dotkey(obj.__class__)]
    for base_class in obj.__class__.__bases__:
        bases.append(dotkey(base_class))
    return bases


class View(dict):
    ''' Provides a virtual view of a Container
    
    Args:
        container (Container): The container
        dotkeys (list[str]): A list of strings specifying the dotkey prefixes
        for the view.
        live (bool): Optional boolean to specify whether the view stays in
        sync with the container. Defaults to settings.LIVE_VIEWS.
        obj (object): Instead of specifying dotkeys, an object can be passed
        in. The bases of the object will be used as dotkeys (after removing
        the top-level key).
        
    Note:
        Views cache values and update them as required for performance. They
        provide both dictionary-style access and attribute-style access to
        values.
        
    Note:
        Subdictionaries are updated from the container. But, any values
        set in subdictionaries on the View do not update those in the
        container.
        
    '''
    def __init__(self, container, dotkeys=None, live=None,
                 obj=None, nesting=True, numpify=None):
        self.__hash = hash(uuid.uuid4())
        self._container = container
        self.dotkeys = []
        self._live = settings.LIVE_VIEWS if live is None else live
        self.dirty = False
        self.nesting = nesting
        self.numpify = numpify
        dotkeys = ['__ROOT__'] if dotkeys is None else dotkeys
        dotkeys = [dotkeys] if isinstance(dotkeys, str) else dotkeys
        dotkeys = get_bases(obj) if obj is not None else dotkeys
        for dotkey in dotkeys:
            self.add_dotkey(dotkey)
    
    @property
    def live(self):
        return self._live
    
    @live.setter
    def live(self, flag):
        assert type(flag) is bool
        self._live = flag
        if self._live:
            self.refresh()
        
    def __hash__(self):
        return self.__hash
        
    def add_dotkey(self, dotkey):
        if dotkey is None or dotkey in ['', '.']:
            dotkey == '__ROOT__'
        self._hook_to_container(dotkey)
    
    def _clash_error(self, key, dotkey):
        raise KeyError('Key "' + str(key) + '" defined in '
                       + dotkey + ' when already present in '
                       + 'one of: "' + '"; "'.join(self.dotkeys) + '"')
        
    def _hook_to_container(self, dotkey):
        self._container.register_observer(dotkey, self)
        self._update(dotkey)
        self.dotkeys.append(dotkey)
        
    def _update(self, dotkey):
        dct = utils.nested(self._container.get_dct(dotkey))
        for key in dct.keys():
            if key in self:
                self._clash_error(key, dotkey)
        for k, v in dct.items():
            self._set_val(k, v)
        
    def refresh(self):
        ''' Refresh all values if required '''
        if not self._live:
            return
        old_keys = self.keys()
        self.clear()
        for dotkey in self.dotkeys:
            self._update(dotkey)
        new_keys = self.keys()
        for old_key in old_keys:
            if old_key not in new_keys:
                delattr(self, old_key)
        self.dirty = False

    def check_refresh(self):
        ''' Refresh if current values only if they are indicated as dirty '''
        if self.dirty:
            self.refresh()
        
    def get(self, key):
        ''' Get a value '''
        return self[key]
        
    def set(self, key, val):
        ''' Set a value 
        
        Note:
            If `live` is true, this method will also set the value in the
            container. The key must already exist.
        '''
        if not self._live:
            self._set_val(key, val)
            return
        container = self._container
        for dotkey in self.dotkeys:
            k = dotkey + '.' + key
            if k in container:
                container.set(k, val)
                return
        raise KeyError(key + ' not found in dotkeys: ' + '; '.join(self.dotkeys))
        
    def __setitem__(self, key, val):
        return self.set(key, val)
    
    def _nested_set(self, key_list, val, dct=None):
        d = self if dct is None else dct
        if len(key_list) > 1:
            root = key_list[0]
            if root not in d:
                d[root] = {}
                setattr(self, root, d[root])
            d_sub = d[root]
            self._nested_set(key_list[1:], val, dct=d_sub)
        else:
            key = key_list[0]
            if d is self:
                self._set_val(key, val)
            else:
                d[key] = val
    
    def _view_update(self, key, val):
        if self._live:
            self._nested_set(key.split('.'), val)
    
    def _set_val(self, key, val):
        if (self.numpify is not None 
            and (self.numpify is True or key in self.numpify)):
                val = utils.numpify(val)
        super().__setitem__(key, val)
        setattr(self, key, val)
    
    def dset(self, dct):
        ''' Set multiple values using a dictionary '''
        for k, v in dct.items():
            self.set(k, v)
            
    @contextmanager
    def context(self, dct):
        ''' A context manager for temporary changes in values 
        
        Args:
            dct (dict): A dictionary of dotkey-value pairs.
        '''
        originals = {k: self[k] for k in dct.keys()}
        self.dset(dct)
        yield self
        self.dset(originals)


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 30 19:31:41 2019

@author: Reuben
"""


import unittest

from vartrix.container import Container

def get_c():
    c = Container()
    c.set('a.b', 5)
    c.set('b.c', 7)
    return c

def get_c2():
    d = {'a': {'b': 1, 'c': 2},
         'd': {'e': 3, 'f': 4, 'g': {'h': 5, 'i': 6}}}
    c = Container(d)
    return c

def get_c2_underscore():
    d = {'a': {'b': 1, 'c': 2},
         'd': {'e': 3, 'f': 4, 'g': {'h': 5, 'i': 6}},
         'd_': {'z': 55}}
    c = Container(d)
    return c

class View_Mock():
    def _view_update(self, key, val):
        self.key = key
        self.val = val
        
    def refresh(self):
        self.refresh_val = True

    def check_refresh(self):
        if self.dirty:
            self.refresh()
        

class Test_Container(unittest.TestCase):
    def test_instantiate(self):
        c = Container()
        self.assertTrue(isinstance(c, dict))

    def test_load_on_init(self):
        dct = {'a.b': 6}
        c = Container(dct)
        self.assertDictEqual(dct, c)

    def test_register_observer(self):
        c = Container()
        b = View_Mock()
        c.register_observer('a', b)
        self.assertIn(b, c._observers['a'])
        
    def test_update_observers(self):
        c = Container()
        b = View_Mock()
        c.register_observer('a', b)
        c.update_observers('a.test_key', 4)
        self.assertEqual(b.key, 'test_key')
        self.assertEqual(b.val, 4)    
    
    def test_dict_set_item(self):
        c = Container()
        c['a'] = 6

    def test_dict_get_item(self):
        c = Container()
        c['a'] = 6
        self.assertEqual(c['a'], 6)
        
    def test_set(self):
        c = Container()
        c.set('a.b', 6)
        self.assertEqual(c['a.b'], 6)

    def test_get(self):
        c = Container()
        c.set('a.b', 6)
        self.assertEqual(c.get('a.b'), 6)

    def test_get_dct_flat(self):
        c = get_c2()
        d = c.get_dct('a')
        expected = {'b': 1, 'c': 2}
        self.assertDictEqual(d, expected)
        
    def test_get_dct_nested(self):
        c = get_c2()
        d = c.get_dct('d')
        expected = {'e': 3, 'f': 4, 'g.h': 5, 'g.i': 6}
        self.assertDictEqual(d, expected)

    def test_get_dct_nested_underscore(self):
        c = get_c2_underscore()
        d = c.get_dct('d')
        expected = {'e': 3, 'f': 4, 'g.h': 5, 'g.i': 6}
        self.assertDictEqual(d, expected)

    def test_get_dct_underscore(self):
        c = get_c2_underscore()
        d = c.get_dct('d_')
        expected = {'z': 55}
        self.assertDictEqual(d, expected)

    def test_lset(self):
        c = Container()
        c.lset(['a', 'b'], 6)
        expected = {'a.b': 6}
        self.assertDictEqual(c, expected)

    def test_lget(self):
        c = Container()
        c.lset(['a', 'b'], 6)
        self.assertEqual(c.lget(['a', 'b']), 6)
        
    def test_dset(self):
        c = Container()
        dct = {'a.b': 6}
        c.dset(dct)
        self.assertDictEqual(c, dct)
        
    def test_nested(self):
        c = get_c()
        ret = c.nested()
        expected = {'a': {'b': 5},
                    'b': {'c': 7}}
        self.assertDictEqual(ret, expected)
        
    def test_load(self):
        c = Container()
        dct = {'a.b': 6}
        c.load(dct)
        self.assertDictEqual(dct, c)
        
    def test_refresh_observer_on_load(self):
        dct = {'a.b': 6}
        c = Container(dct)
        b = View_Mock()
        c.register_observer('a', b)
        dct2 = {'a.b': 7,
                'a.e': 45}
        c.load(dct2)
        self.assertEqual(b.refresh_val, True)

    def test_combine(self):
        d1 = {'b': 5}
        d2 = {'d': 7}
        c1 = Container(d1)
        c2 = Container(d2)
        c = Container.combine({'c1': c1, 'c2': c2})
        expected = {'c1.b': 5,
                    'c2.d': 7}
        self.assertEqual(c, expected)
        
    def test_merge(self):
        d1 = {'a.b': 5}
        d2 = {'c.d': 7}
        c1 = Container(d1)
        b1 = View_Mock()
        c1.register_observer('a', b1)
        c2 = Container(d2)
        b2 = View_Mock()
        c2.register_observer('c', b2)
        c = Container.merge([c1, c2])
        expected = {'a.b': 5,
                    'c.d': 7}
        self.assertEqual(c, expected)
        self.assertIn(b1, c._observers['a'])
        self.assertIn(b2, c._observers['c'])
        
    def test_context(self):
        dct = {'a.b': 6, 'a.c': 7}
        c = Container(dct)
        new1 = {'a.b': 66}
        new2 = {'a.c': 77}
        with c.context(new1):
            expected1 = {'a.b': 66, 'a.c': 7}
            self.assertDictEqual(c, expected1)
            with c.context(new2):
                expected2 = {'a.b': 66, 'a.c': 77}
                self.assertDictEqual(c, expected2)
            self.assertDictEqual(c, expected1)
        self.assertDictEqual(dct, c)
        
    def test_reset(self):
        c = get_c2()
        c.set('z.c', 7)
        c.reset()
        dct = {'a.b': 1, 'a.c': 2,
             'd.e': 3, 'd.f': 4, 'd.g.h': 5, 'd.g.i': 6}
        self.assertDictEqual(dct, c)


# -*- coding: utf-8 -*-
"""
Created on Sat Sep 14 11:36:50 2019

@author: Reuben

This module contains helper classes and functions. 

"""

import inspect, sys
import numpy as np

class Factory():
    ''' Instantiate module classes based on container key 
    
    Args:
        name (str): The module name (often `__name__`)
        container (Container): The container instance
        dotkey (str): The dotkey of the class name.
        build_function (func): OPTIONAL: A function that takes the new
        instance object. If provided, the return value of the Factory is
        the return value of the function.
        
    Note:
        The class name pointed to by the dotkey must exist in the module,
        or a KeyError will be raised.
    '''
    
    def __init__(self, name, container, dotkey, build_function=None):
        self.container = container
        self.name = name
        self.dotkey = dotkey
        self.build_function = build_function
    
    def new(self):
        ''' Create a new instance based on the current dotkey value '''
        cls_name = self.container[self.dotkey]
        clsmembers = inspect.getmembers(sys.modules[self.name],
                                        inspect.isclass)
        clsdct = {t[0]: t[1] for t in clsmembers}
        try:
            c = clsdct[cls_name]
        except KeyError:
            raise KeyError('Class ' + str(cls_name) + ' not found in module '
                           + str(self.name))
        if self.build_function is None:
            return c()
        else:
            obj = c()
            return self.build_function(obj)
        
        
def _nest(key_list, val, dct=None):
    dct = {} if dct is None else dct
    if len(key_list) == 1:
        dct[key_list[0]] = val
    else:
        if key_list[0] not in dct:
            dct[key_list[0]] = {}
        _nest(key_list[1:], val, dct[key_list[0]])
    return dct

def nested(dct):
    ''' Create a nested dictionary representation of a dotkey flat dictionary '''
    out_dct = {}
    for dotkey, val in dct.items():
        key_list = dotkey.split('.')
        _nest(key_list, val, out_dct)
    return out_dct

def numpify(obj):
    if isinstance(obj, list):
        return np.array(obj)
    else:
        return obj
    
def denumpify(obj):
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    else:
        return obj


# -*- coding: utf-8 -*-
"""
Created on Mon Sep  9 17:05:43 2019

@author: Reuben
"""

import ruamel.yaml as yml

class Manager():
    default_handler = 'yaml'
    
    def __init__(self):
        self.handlers = {'yaml': Yaml()}
        self.specified = None
        
    def add_handler(self, key, handler):
        self.handlers[key] = handler
        
    def specify(self, key):
        self.specified = key
        
    def load(self, source, handler=None, **kwargs):
        if handler is not None:
            return self.handlers[handler].load(source, **kwargs)
        if self.specified is not None:
            return self.handlers[self.specified].load(source, **kwargs)
        for k, handler in self.handlers.items():
            if handler.suitable(source, **kwargs):
                return handler.load(source, **kwargs)
        f = source + '.yaml'
        return self.handlers[self.default_handler].load(f, **kwargs)

    def save(self, dct, target, handler=None, **kwargs):
        if handler is not None:
            return self.handlers[handler].save(target, **kwargs)
        if self.specified is not None:
            return self.handlers[self.specified].save(dct, target, **kwargs)
        for k, handler in self.handlers.items():
            if handler.suitable(target, **kwargs):
                return handler.save(dct, target, **kwargs)
        f = target + '.yaml'
        return self.handlers[self.default_handler].save(dct, f, **kwargs)
        

class Yaml():
    def suitable(self, fname, **kwargs):
        test_1 = fname.endswith('.yml')
        test_2 = fname.endswith('.yaml')
        return test_1 or test_2
        
    def load(self, fname, **kwargs):
        with open(fname, mode='r') as f:
            dct = yml.safe_load(f)
        return dct
    
    def save(self, dct, fname, **kwargs):
        with open(fname, mode='w') as f:
            y = yml.YAML()
            y.dump(dict(dct), f)
            
            
manager = Manager()
load = manager.load
save = manager.save



# -*- coding: utf-8 -*-
"""
Created on Thu Sep  5 11:14:59 2019

@author: Reuben
"""

from .container import Container

class Name_Space(dict):
    def __init__(self, obj_cls=Container):
        self.i = 0
        self.obj_cls = obj_cls
        self._active_container = None
    
    def __missing__(self, key):
        return self.create(key)
    
    def create(self, key, dct=None):
        new = self.obj_cls(name=key, dct=dct)
        self[key] = new
        return new
    
    def get(self, key=None, dct=None):
        if key is None:
            key = self.i
            self.i += 1
        if key not in self:
            return self.create(key, dct)
        return self[key]

    def set_active(self, key):
        if key in self or key is None:
            self._active_container = key
        else:
            raise KeyError(key + ' not found in name space.')
    
    def active_container(self):
        return self[self._active_container]
    
    def duplicate(self, key, new_key):
        self[new_key] = self[key].copy()
        return self[new_key]


default = Name_Space()

def get_container(name=None):
    return default.get(name)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep  9 15:36:55 2019

@author: Reuben
"""

import unittest

from vartrix import get_container, Name_Space, Container

class Test_Name_Space(unittest.TestCase):
    def test_init(self):
        ns = Name_Space()
        
    def test_create(self):
        ns = Name_Space(obj_cls=Container)
        container = ns.create('test')
        self.assertTrue(isinstance(container, Container))
        self.assertEqual(container.name, 'test')
        
    def test_create_dct(self):
        ns = Name_Space(obj_cls=Container)
        dct = {'a.b': 5, 'c.d': 7}
        container = ns.create('test', dct=dct)
        self.assertTrue(isinstance(container, Container))
        self.assertEqual(container.name, 'test')
        self.assertDictEqual(container, dct)
    
    def test_get_create(self):
        ns = Name_Space(obj_cls=Container)
        container = ns.get('test')
        self.assertTrue(isinstance(container, Container))
        self.assertEqual(container.name, 'test')
        
    def test_get_dct_create(self):
        ns = Name_Space(obj_cls=Container)
        dct = {'a.b': 5, 'c.d': 7}
        container = ns.get('test', dct=dct)
        self.assertTrue(isinstance(container, Container))
        self.assertEqual(container.name, 'test')
        self.assertDictEqual(container, dct)

    def test_get_repeated(self):
        ns = Name_Space(obj_cls=Container)
        container1 = ns.get('test')
        container2 = ns.get('test')
        self.assertTrue(container1 is container2)

class Test_get_container(unittest.TestCase):
    def test_get_container_create(self):
        container = get_container('test')
        self.assertTrue(isinstance(container, Container))
        self.assertEqual(container.name, 'test')

    def test_get_container_repeated(self):
        container1 = get_container('test2')
        container2 = get_container('test2')
        self.assertTrue(container1 is container2)




# -*- coding: utf-8 -*-
"""
Created on Wed Sep  4 08:47:20 2019

@author: Reuben
"""

import os
import unittest

import ruamel.yaml as yml

from vartrix.container import Container
from vartrix.view import View
from vartrix.namespace import Name_Space
from vartrix import automate

base = {'alias.one': 5,
        'alias.two': 7,
        'alias.three': 11,
        'alias.four': 17}

namespace = Name_Space(Container)
container = namespace.get('test', base)

def get_c():
    f = namespace.get('test', base)
    return f

def set_csv_root():
    automate.set_root(os.path.dirname(os.path.abspath(__file__)))

def get_fname():
    root = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(root, 'automation_sets.yml')
    return fname    

def get_test_data():
    fname = get_fname()
    set_csv_root()
    with open(fname) as f:
        sets = yml.safe_load(f)
    return sets


class Automated():
    def __init__(self, set_name):
        dotkeys = ['alias']
        self.params = View(namespace.get('test'), dotkeys)
        self.set_name = set_name
        self.prepare_sequence_names = []
        self.finish_sequence_names = []
        self.prepare_method_names = []
        self.finish_method_names = []
        self.method_a_calls = []
        self.method_b_calls = []
        self.method_c_calls = []
        self.method_d_calls = []
        def get_methods_dct():
            return {'method_a': [], 'method_b': [], 'method_c': [], 'method_d': []}
        def get_dct():
            return {'seq_1': get_methods_dct(),
                    'seq_2': get_methods_dct(),
                    'seq_3': get_methods_dct(),
                    'seq_4': get_methods_dct()}
        self.alias_one_history = get_dct()
        self.alias_two_history = get_dct()
        self.alias_three_history = get_dct()
        self.alias_four_history = get_dct()
        
    def prepare(self):
        self.prepare = True

    def prepare_sequence(self, seq_name):
        self.prepare_sequence_names.append(seq_name)

    def prepare_method(self, method_name):
        self.prepare_method_names.append(method_name)

    def append_history(self, seq_name, method_name):
        self.alias_one_history[seq_name][method_name].append(self.params.one)
        self.alias_two_history[seq_name][method_name].append(self.params.two)
        self.alias_three_history[seq_name][method_name].append(self.params.three)
        self.alias_four_history[seq_name][method_name].append(self.params.four)

    def method_a(self, seq_name, val_dct, label_dct):
        self.method_a_calls.append({'val_dct': val_dct,
                                    'label_dct': label_dct})
        self.append_history(seq_name, 'method_a')

    def method_b(self, seq_name, val_dct, label_dct):
        self.method_b_calls.append({'val_dct': val_dct,
                                    'label_dct': label_dct})
        self.append_history(seq_name, 'method_b')

    def method_c(self, seq_name, val_dct, label_dct):
        self.method_c_calls.append({'val_dct': val_dct,
                                    'label_dct': label_dct})
        self.append_history(seq_name, 'method_c')
    
    def method_d(self, seq_name, val_dct, label_dct):
        self.method_d_calls.append({'val_dct': val_dct,
                                    'label_dct': label_dct})
        self.append_history(seq_name, 'method_d')

    def finish_method(self, method_name):
        self.finish_method_names.append(method_name)

    def finish_sequence(self, seq_name):
        self.finish_sequence_names.append(seq_name)

    def finish(self):
        self.finish = True


class Test_Automator(unittest.TestCase):
    def test_run_seq_1(self):
        container = get_c()
        fname = get_fname()
        set_csv_root()
        a = automate.Automator(container, fname)
        automated = Automated('set_1')
        a.run('set_1', automated)
        
        expected_one = {
            'seq_1': {
                'method_a': [0, 2, 3, 4],
                'method_b': [],
                'method_c': [],
                'method_d': []},
            'seq_2': {
                'method_a': [5, 5, 5, 5, 5, 5],
                'method_b': [0, 0, 0, 2, 2, 2, 3, 3, 3, 4, 4, 4],
                'method_c': [],
                'method_d': []},
            'seq_3': {
                'method_a': [],
                'method_b': [],
                'method_c': [0, 0, 2, 2, 3, 3, 4, 4],
                'method_d': []},
            'seq_4': {
                'method_a': [],
                'method_b': [],
                'method_c': [11, 13, 14, 11, 13, 14],
                'method_d': []}}
        self.assertDictEqual(automated.alias_one_history, expected_one)

        expected_two = {'seq_1': {'method_a': [7, 7, 7, 7], 'method_b': [], 'method_c': [], 'method_d': []}, 'seq_2': {'method_a': [2, 2, 3, 3, 4, 4], 'method_b': [2, 3, 4, 2, 3, 4, 2, 3, 4, 2, 3, 4], 'method_c': [], 'method_d': []}, 'seq_3': {'method_a': [], 'method_b': [], 'method_c': [7, 7, 7, 7, 7, 7, 7, 7], 'method_d': []}, 'seq_4': {'method_a': [], 'method_b': [], 'method_c': [12, 14, 16, 12, 14, 16], 'method_d': []}}
        self.assertDictEqual(automated.alias_two_history, expected_two)

        expected_three = {'seq_1': {'method_a': [11, 11, 11, 11], 'method_b': [], 'method_c': [], 'method_d': []}, 'seq_2': {'method_a': [6, 6, 7, 7, 8, 8], 'method_b': [6, 7, 8, 6, 7, 8, 6, 7, 8, 6, 7, 8], 'method_c': [], 'method_d': []}, 'seq_3': {'method_a': [], 'method_b': [], 'method_c': [11, 11, 11, 11, 11, 11, 11, 11], 'method_d': []}, 'seq_4': {'method_a': [], 'method_b': [], 'method_c': [11, 11, 11, 11, 11, 11], 'method_d': []}}
        self.assertDictEqual(automated.alias_three_history, expected_three)

        expected_four = {'seq_1': {'method_a': [17, 17, 17, 17], 'method_b': [], 'method_c': [], 'method_d': []}, 'seq_2': {'method_a': [4, 5, 4, 5, 4, 5], 'method_b': [17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17], 'method_c': [], 'method_d': []}, 'seq_3': {'method_a': [], 'method_b': [], 'method_c': [4, 5, 4, 5, 4, 5, 4, 5], 'method_d': []}, 'seq_4': {'method_a': [], 'method_b': [], 'method_c': [4, 4, 4, 5, 5, 5], 'method_d': []}}
        self.assertDictEqual(automated.alias_four_history, expected_four)



class Test_Aliases(unittest.TestCase):
    pass


class Test_Vector(unittest.TestCase):
    def get_vec_1(self):
        d = {'labels': ['a', 'b'],
             'alias_1': [1, 2],
             'alias_2': [3, 4]}
        return automate.Vector_Factory.new(d, 'vec_1')
    
    def test_init_labelled(self):
        v = self.get_vec_1()
        expected = [{'alias_1': 1, 'alias_2': 3}, {'alias_1': 2, 'alias_2': 4}]
        self.assertListEqual(v.data, expected)
        self.assertListEqual(v.labels, ['a', 'b'])
        
    def test_get_lst_labelled(self):
        v = self.get_vec_1()
        lst = v.get_lst()
        self.assertListEqual(lst, [{'alias_1': 1, 'alias_2': 3},
                                   {'alias_1': 2, 'alias_2': 4}])

    def test_get_label_lst_labelled(self):
        v = self.get_vec_1()
        lst = v.get_label_lst()
        self.assertListEqual(lst, [{'vec_1': 'a'},
                                   {'vec_1': 'b'}])

    def get_vec_unlabelled(self):
        d = {'alias_1': [1, 2],
             'alias_2': [3, 4]}
        return automate.Vector_Factory.new(d, 'vec_unlabelled')
    
    def test_init_unlabelled(self):
        v = self.get_vec_unlabelled()
        expected = [{'alias_1': 1, 'alias_2': 3}, {'alias_1': 2, 'alias_2': 4}]
        self.assertListEqual(v.data, expected)
        self.assertListEqual(v.labels, [0, 1])
        
    def test_get_lst_unlabelled(self):
        v = self.get_vec_unlabelled()
        lst = v.get_lst()
        self.assertListEqual(lst, [{'alias_1': 1, 'alias_2': 3},
                                   {'alias_1': 2, 'alias_2': 4}])

    def test_get_label_lst_unlabelled(self):
        v = self.get_vec_unlabelled()
        lst = v.get_label_lst()
        self.assertListEqual(lst, [{'vec_unlabelled': 0},
                                   {'vec_unlabelled': 1}])


    def get_vec_2(self):
        d = {'labels': ['c', 'd'],
             'alias_3': [5, 6],
             'alias_4': [7, 8]}
        return automate.Vector_Factory.new(d, 'vec_2')
    
    def test_get_lst_nested(self):
        v = self.get_vec_1()
        v2 = self.get_vec_2()
        v.set_child(v2)
        lst = v.get_lst()
        expected = [
            {'alias_1': 1, 'alias_2': 3, 'alias_3': 5, 'alias_4': 7},
            {'alias_1': 1, 'alias_2': 3, 'alias_3': 6, 'alias_4': 8},
            {'alias_1': 2, 'alias_2': 4, 'alias_3': 5, 'alias_4': 7},
            {'alias_1': 2, 'alias_2': 4, 'alias_3': 6, 'alias_4': 8}]
        self.assertListEqual(lst, expected)

    def test_get_label_lst_nested(self):
        v = self.get_vec_1()
        v2 = self.get_vec_2()
        v.set_child(v2)
        lst = v.get_label_lst()
        expected = [{'vec_1': 'a', 'vec_2': 'c'},
                    {'vec_1': 'a', 'vec_2': 'd'},
                    {'vec_1': 'b', 'vec_2': 'c'},
                    {'vec_1': 'b', 'vec_2': 'd'}]
        self.assertListEqual(lst, expected)

    def test_csv(self):
        automate.root = os.path.dirname(os.path.abspath(__file__))
        dct = {'style': 'csv', 'filename': 'test_data.csv'}
        v = automate.Vector_Factory.new(dct, 'csv_vec')
        lst = v.get_lst()
        expected = [{'alias_1': 2, 'alias_2': 3},
                    {'alias_1': 5, 'alias_2': 7}]
        self.assertListEqual(lst, expected)
        expected = [{'csv_vec': 'row_one'},
                    {'csv_vec': 'row_two'}]
        lst = v.get_label_lst()
        


class Test_List_Vector(unittest.TestCase):
    def get_d_1(self):
        d = {'alias_1': [1, 2],
             'alias_2': [3, 4]}
        return d
    
    def test_transpose_dict(self):
        d = self.get_d_1()
        v = automate.List_Vector('test_name')
        d_test = v.transpose_dict(d)
        expected = [{'alias_1': 1, 'alias_2': 3}, {'alias_1': 2, 'alias_2': 4}]
        self.assertListEqual(d_test, expected)
    
class Test_Dict_List_Vector(unittest.TestCase):

    def test_int(self):
        d = {'a':  {'b': 1}}
        v = automate.Dict_List_Vector('test_name')
        labels, d = v.setup(d)
        expected = [{'b': 1}]
        self.assertListEqual(d, expected)
        self.assertListEqual(labels, ['a'])
    
    def test_vector(self):
        d = {'a':  {'b': [ 1,  2,  3]}}
        v = automate.Dict_List_Vector('test_name')
        labels, d = v.setup(d)
        expected = [{'b': [1, 2, 3]}]
        self.assertListEqual(d, expected)
        self.assertListEqual(labels, ['a'])
            
    
class Test_Vectors(unittest.TestCase):
    def make_vectors(self):
        d = {'vec_1': {'labels': ['a', 'b'],
                       'alias_1': [1, 2],
                       'alias_2': [3, 4]},
             'vec_2': {'labels': ['c', 'd'],
                       'alias_3': [5, 6],
                       'alias_4': [7, 8]}}
        v = automate.Vectors(d)    
        return v
    
    def test_loop(self):
        v = self.make_vectors()
        val_lst, label_lst = v.loop(['vec_1', 'vec_2'])
        self.assertListEqual(val_lst, [{'alias_1': 1, 'alias_2': 3, 'alias_3': 5, 'alias_4': 7},
                                       {'alias_1': 1, 'alias_2': 3, 'alias_3': 6, 'alias_4': 8},
                                       {'alias_1': 2, 'alias_2': 4, 'alias_3': 5, 'alias_4': 7},
                                       {'alias_1': 2, 'alias_2': 4, 'alias_3': 6, 'alias_4': 8}])
        self.assertListEqual(label_lst, [{'vec_1': 'a', 'vec_2': 'c'},
                                         {'vec_1': 'a', 'vec_2': 'd'},
                                         {'vec_1': 'b', 'vec_2': 'c'},
                                         {'vec_1': 'b', 'vec_2': 'd'}])


class Test_Sequencer(unittest.TestCase):
    def get_s(self):
        all_data = get_test_data()
        data = all_data['set_1']
        s = automate.Sequencer(data['sequences'],
                               all_data['aliases'],
                               data['vectors'])
        return s
    
    def test_seq_1(self):
        s = self.get_s()
        sd = s.sequence('seq_1')
        expected = {'method_a': {'val_list':
                                    [{'alias.one': 0},
                                     {'alias.one': 2},
                                     {'alias.one': 3},
                                     {'alias.one': 4}],
                                'label_lst': 
                                    [{'vec_1': 0},
                                     {'vec_1': 2},
                                     {'vec_1': 3},
                                     {'vec_1': 4}]}}
        self.assertDictEqual(sd, expected)
        
    def test_seq_2(self):
        s = self.get_s()
        sd = s.sequence('seq_2')
        expected = {
            'method_b': {
                'val_list': [
                    {'alias.one': 0, 'alias.two': 2, 'alias.three': 6},
                    {'alias.one': 0, 'alias.two': 3, 'alias.three': 7},
                    {'alias.one': 0, 'alias.two': 4, 'alias.three': 8},
                    {'alias.one': 2, 'alias.two': 2, 'alias.three': 6},
                    {'alias.one': 2, 'alias.two': 3, 'alias.three': 7},
                    {'alias.one': 2, 'alias.two': 4, 'alias.three': 8},
                    {'alias.one': 3, 'alias.two': 2, 'alias.three': 6},
                    {'alias.one': 3, 'alias.two': 3, 'alias.three': 7},
                    {'alias.one': 3, 'alias.two': 4, 'alias.three': 8},
                    {'alias.one': 4, 'alias.two': 2, 'alias.three': 6},
                    {'alias.one': 4, 'alias.two': 3, 'alias.three': 7},
                    {'alias.one': 4, 'alias.two': 4, 'alias.three': 8}],
                'label_lst': [
                    {'vec_1': 0, 'vec_2': 'a'},
                    {'vec_1': 0, 'vec_2': 'b'},
                    {'vec_1': 0, 'vec_2': 'c'},
                    {'vec_1': 2, 'vec_2': 'a'},
                    {'vec_1': 2, 'vec_2': 'b'},
                    {'vec_1': 2, 'vec_2': 'c'},
                    {'vec_1': 3, 'vec_2': 'a'},
                    {'vec_1': 3, 'vec_2': 'b'},
                    {'vec_1': 3, 'vec_2': 'c'},
                    {'vec_1': 4, 'vec_2': 'a'},
                    {'vec_1': 4, 'vec_2': 'b'},
                    {'vec_1': 4, 'vec_2': 'c'}]}, 
            'method_a': {'val_list': [
                    {'alias.two': 2, 'alias.three': 6, 'alias.four': 4},
                    {'alias.two': 2, 'alias.three': 6, 'alias.four': 5},
                    {'alias.two': 3, 'alias.three': 7, 'alias.four': 4},
                    {'alias.two': 3, 'alias.three': 7, 'alias.four': 5},
                    {'alias.two': 4, 'alias.three': 8, 'alias.four': 4},
                    {'alias.two': 4, 'alias.three': 8, 'alias.four': 5}],
                'label_lst': [
                    {'vec_2': 'a', 'vec_3': 'four'},
                    {'vec_2': 'a', 'vec_3': 'five'},
                    {'vec_2': 'b', 'vec_3': 'four'},
                    {'vec_2': 'b', 'vec_3': 'five'},
                    {'vec_2': 'c', 'vec_3': 'four'},
                    {'vec_2': 'c', 'vec_3': 'five'}]}}
        self.assertDictEqual(sd, expected)
        
    
    def test_seq_3(self):
        s = self.get_s()
        sd = s.sequence('seq_3')
        expected = {
            'method_c': {
                'val_list': [
                    {'alias.one': 0, 'alias.four': 4},
                    {'alias.one': 0, 'alias.four': 5},
                    {'alias.one': 2, 'alias.four': 4},
                    {'alias.one': 2, 'alias.four': 5},
                    {'alias.one': 3, 'alias.four': 4},
                    {'alias.one': 3, 'alias.four': 5},
                    {'alias.one': 4, 'alias.four': 4},
                    {'alias.one': 4, 'alias.four': 5}],
                'label_lst': [
                    {'vec_1': 0, 'vec_3': 'four'},
                    {'vec_1': 0, 'vec_3': 'five'},
                    {'vec_1': 2, 'vec_3': 'four'},
                    {'vec_1': 2, 'vec_3': 'five'},
                    {'vec_1': 3, 'vec_3': 'four'},
                    {'vec_1': 3, 'vec_3': 'five'},
                    {'vec_1': 4, 'vec_3': 'four'},
                    {'vec_1': 4, 'vec_3': 'five'}]}}
        self.assertDictEqual(sd, expected)
        

    def test_seq_4(self):
        s = self.get_s()
        sd = s.sequence('seq_4')
        expected = {
            'method_c': {
                'val_list': [
                    {'alias.four': 4, 'alias.one': 11, 'alias.two': 12},
                    {'alias.four': 4, 'alias.one': 13, 'alias.two': 14},
                    {'alias.four': 4, 'alias.one': 14, 'alias.two': 16},
                    {'alias.four': 5, 'alias.one': 11, 'alias.two': 12},
                    {'alias.four': 5, 'alias.one': 13, 'alias.two': 14},
                    {'alias.four': 5, 'alias.one': 14, 'alias.two': 16}],
                'label_lst': [
                    {'vec_3': 'four', 'vec_4': 'a'},
                    {'vec_3': 'four', 'vec_4': 'b'},
                    {'vec_3': 'four', 'vec_4': 'c'},
                    {'vec_3': 'five', 'vec_4': 'a'},
                    {'vec_3': 'five', 'vec_4': 'b'},
                    {'vec_3': 'five', 'vec_4': 'c'}]}}
        self.assertDictEqual(sd, expected)
        
    def get_s2(self):
        all_data = get_test_data()
        data = all_data['set_2']
        d = data['vectors']
        d.update(data['constants'])
        s = automate.Sequencer(data['sequences'],
                               all_data['aliases'],
                               d)
        return s
    
    def test_seq_10(self):
        s = self.get_s2()
        sd = s.sequence('seq_10')
        expected = {
            'method_a': {
                'val_list': [
                    {'alias.one': 5, 'alias.two': 7}],
                'label_lst': [
                    {'const_1': 5, 'const_2': 7}]}}
        self.assertDictEqual(sd, expected)
        
    def test_seq_11(self):
        s = self.get_s2()
        sd = s.sequence('seq_11')
        expected = {
            'method_a': {
                'val_list': [
                    {'alias.one': 5, 'alias.three': 1},
                    {'alias.one': 5, 'alias.three': 2},
                    {'alias.one': 5, 'alias.three': 3}],
                'label_lst': [
                    {'const_1': 5, 'vec_1': 1},
                    {'const_1': 5, 'vec_1': 2},
                    {'const_1': 5, 'vec_1': 3}]}}
        self.assertDictEqual(sd, expected)


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 19 15:02:12 2019

@author: Reuben
"""

import setuptools

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="vartrix",
    version="0.0.8",
    author="Reuben Rusk",
    description="Easily manange and automate variables and parameters.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pythoro/vartrix.git",
    download_url="https://github.com/pythoro/vartrix/archive/v0.0.8.zip",
    packages=['vartrix'],
    keywords=['PARAMETERS', 'VARIABLES', 'PARAMETRIC', 'AUTOMATION', 'AUTOMATE'],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    install_requires=[],
)


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 26 16:12:03 2019

@author: Reuben
"""

from . import utils
from . import settings
from . import persist
from . import container, view
from . import automate
from . import namespace


from .persist import load, save
from .namespace import Name_Space, get_container
from .container import Container
from .view import View
from .automate import Automator
from .utils import Factory




# -*- coding: utf-8 -*-
"""
Created on Tue Aug 27 21:11:42 2019

@author: Reuben
"""

import unittest

from vartrix.container import Container
from vartrix.view import View
from vartrix import settings

def get_c():
    c = Container()
    c.set('a.b', 5)
    c.set('b.c', 7)
    return c

class Test_View(unittest.TestCase):
    def test_init(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        self.assertTrue(isinstance(v, dict))
        
    def test_init_none(self):
        c = get_c()
        v = View(c, dotkeys=None)
        expected = {'a': {'b': 5}, 'b': {'c': 7}}
        self.assertDictEqual(v, expected)

    def test_init_empty(self):
        c = get_c()
        v = View(c, dotkeys=[''])
        expected = {'a': {'b': 5}, 'b': {'c': 7}}
        self.assertDictEqual(v, expected)

    def test_init_dot(self):
        c = get_c()
        v = View(c, dotkeys=['.'])
        expected = {'a': {'b': 5}, 'b': {'c': 7}}
        self.assertDictEqual(v, expected)
    
    def test_get(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        res = v['b']
        expected = 5
        self.assertEqual(res, expected)

    def test_get_attr(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        res = v.b
        expected = 5
        self.assertEqual(res, expected)
        
    def test_refresh(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        v.clear()
        v.refresh()
        v.refresh()
        res = v['b']
        expected = 5
        self.assertEqual(res, expected)
               
    def test_refresh_attr(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        del v.b
        v.refresh()
        res = v.b
        expected = 5
        self.assertEqual(res, expected)   
                
    def test_set(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        v['b'] = 99
        res = v['b']
        expected = 99
        self.assertEqual(res, expected)
        res2 = c['a.b']
        self.assertEqual(res2, expected)

    def test_nested_set(self):
        c = get_c()
        c['a.g.h'] = 55
        dotkeys = ['a']
        v = View(c, dotkeys)
        v['g.h'] = 99
        res = v['g']
        expected = {'h': 99}
        self.assertDictEqual(res, expected)
        
        
    def test_set_remote(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        c['a.b'] = 99
        res = v['b']
        expected = 99
        self.assertEqual(res, expected)
        res2 = c['a.b']
        self.assertEqual(res2, expected)
        
    def test_new_key(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        c['a.new'] = 99
        res = v['new']
        expected = 99
        self.assertEqual(res, expected)

    def test_live_remote(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        v.live = False
        c['a.b'] = 99
        res = v['b']
        expected = 5
        self.assertEqual(res, expected)
        res2 = c['a.b']
        expected2 = 99
        self.assertEqual(res2, expected2)
        v.live = True
        res3 = v['b']
        expected3 = 99
        self.assertEqual(res3, expected3)
        
    def test_live_local(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        v.live = False
        v['b'] = 99
        res = c['a.b']
        expected = 5
        self.assertEqual(res, expected)
        res2 = v['b']
        expected2 = 99
        self.assertEqual(res2, expected2)
        v.live = True
        res3 = v['b']
        expected3 = 5
        self.assertEqual(res3, expected3)

    def test_live_settings(self):
        c = get_c()
        dotkeys = ['a']
        settings.LIVE_VIEWS = False
        v = View(c, dotkeys)
        # v.live = False
        settings.LIVE_VIEWS = True
        c['a.b'] = 99
        res = v['b']
        expected = 5
        self.assertEqual(res, expected)
        res2 = c['a.b']
        expected2 = 99
        self.assertEqual(res2, expected2)
        v.live = True
        res3 = v['b']
        expected3 = 99
        self.assertEqual(res3, expected3)
        
    def test_context(self):
        c = get_c()
        dotkeys = ['a']
        v = View(c, dotkeys)
        with v.context({'b': 77}):
            self.assertEqual(v['b'], 77)
        self.assertEqual(v['b'], 5)
        


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 26 22:12:43 2019

@author: Reuben
"""

import os
import pandas as pd

from . import persist, settings

root = os.path.dirname(os.path.abspath(__file__))

def set_root(new):
    global root
    root = new
    

def safe_call(method_name, obj, *args, **kwargs):
    try:
        method = getattr(obj, method_name)
        method(*args, **kwargs)
    except AttributeError:
        pass

class Automator():
    def __init__(self, container, fname):
        self.container = container
        self.sets = persist.load(fname)
            
    def show(self, method_name, seq_name, label_dct):
        print('Executing "' + method_name + '" method in "' + seq_name +
              '" sequence with:')
        for vec_name, item_name in label_dct.items():
            print('   ' + (vec_name + ': ').ljust(25) + str(item_name))
        
    def run(self, set_name, obj):
        ''' Run an automation set 
        
        Args:
            set_name (str): The name of the set to run
            obj (object): The automated class instance.
        '''
        data = self.sets[set_name]
        vec_data = {}
        if 'vectors' in data:
            vec_data.update(data['vectors'])
        if 'constants' in data:
            vec_data.update(data['constants'])
        s = Sequencer(data['sequences'], self.sets['aliases'], vec_data)
        safe_call('prepare', obj)
        for seq_name, seq_dct in s.all_sequences().items():
            self.execute_sequence(seq_name, seq_dct, obj)
        safe_call('finish', obj)

    def execute_sequence(self, seq_name, seq_dct, obj):
        safe_call('prepare_sequence', obj, seq_name)
        for method_name, dct in seq_dct.items():
            self.execute_method(method_name, seq_name,
                                dct['val_list'],
                                dct['label_lst'],
                                obj)
        safe_call('finish_sequence', obj, seq_name)

    def execute_method(self, method_name, seq_name, val_list, label_lst, obj):
        container = self.container
        safe_call('prepare_method', obj, method_name)
        method = getattr(obj, method_name)
        with container.context(val_list[0]):
            for val_dct, label_dct in zip(val_list, label_lst):
                container.dset(val_dct)
                if settings.PRINT_UPDATES:
                    self.show(method_name, seq_name, label_dct)
                method(seq_name, val_dct, label_dct)
        safe_call('finish_method', obj, method_name)
        
    
class Aliases(dict):
       
    def translate(self, dct):
        out = {}
        for k, v in dct.items():
            if k in self:
                out[self[k]] = v
            else:
                out[k] = v
        return out
    
    
class Vector():
    ''' Subclass for different entry formats '''
    def __init__(self, name):
        self.name =  name
        self.i = 0
        self.child = None
        
    def initialise(self, data):
        self.labels, self.data = self.setup(data)
        self.n = len(self.labels)
    
    def setup(self, data):
        raise NotImplementedError
        
    def set_child(self, vector):
        self.child = vector
        
    def get_lst(self, lst=None, d=None):
        lst = [] if lst is None else lst
        d = {} if d is None else d
        for dct in self.data:
            d.update(dct)
            if self.child is not None:
                self.child.get_lst(lst, d)
            else:
                lst.append(d.copy())
        return lst
    
    def get_label_lst(self, lst=None, d=None):
        lst = [] if lst is None else lst
        d = {} if d is None else d
        for label in self.labels:
            d[self.name] = label
            if self.child is not None:
                self.child.get_label_lst(lst, d)
            else:
                lst.append(d.copy())
        return lst

class List_Vector(Vector):
    def transpose_dict(self, dct):
        n = len(list(dct.values())[0])
        out = [{} for i in range(n)]
        for i in range(n):
            for k, v in dct.items():
                out[i][k] = v[i]
        return out
    
    def setup(self, data):
        def default_labels(d):
            for v in dct.values():
                if len(d) == 1:
                    return v
                else:
                    n = len(v)
                    return [i for i in range(n)]
        dct = data.copy()
        labels = dct.pop('labels', default_labels(dct))
        return labels, self.transpose_dict(dct)    

class Dict_List_Vector(Vector):
    def setup(self, data):
        labels = []
        d = []
        for k, dct in data.items():
            labels.append(k)
            d.append(dct)
        return labels, d

class Constant(Vector):
    def setup(self, data):
        k, v = data.copy().popitem()
        labels = [v]
        d = [data.copy()]
        return labels, d

class Csv_File(Vector):
    def setup(self, data):
        d = []
        full_filename = os.path.join(root, data['filename'])
        df = pd.read_csv(full_filename, index_col=0)
        labels = list(df.index)
        for index, row in df.iterrows():
            d.append(row.to_dict())
        return labels, d

    
class Vector_Factory():
    styles = {'value_lists': List_Vector,
              'value_dictionaries': Dict_List_Vector,
              'csv': Csv_File,
              'constant': Constant}
    default_style = 'value_lists'

    @classmethod
    def set_style(cls, style_name, vec_cls):
        cls.styles[style_name] = vec_cls
    
    @classmethod
    def new(cls, data, name):
        if 'style' not in data:
            vec_cls = cls.styles[cls.default_style]
        else:
            vec_cls = cls.styles[data['style']]
        if len(data) == 1:
            for k, v in data.items():
                if not isinstance(v, (list, dict)):
                    vec_cls = cls.styles['constant']
        v = vec_cls(name)
        d = data.copy()
        try:
            d.pop('style')
        except KeyError:
            pass
        v.initialise(d)
        return v
            
    
class Vectors():
    def __init__(self, data):
        self.data = data
    
    def loop(self, v_names):
        root = None
        vectors = {k: Vector_Factory.new(v, k) for k, v in self.data.items()}
        for v_name in v_names:
            if root is None:
                root = vectors[v_name]
                current = vectors[v_name]
            else:
                current.set_child(vectors[v_name])
                current = vectors[v_name]
        return root.get_lst(), root.get_label_lst()
        
        
class Sequencer():
    def __init__(self, sequences, aliases, vectors):
        self.sequences = sequences
        self.aliases = Aliases(aliases)
        self.vectors = Vectors(vectors)
        
    def sequence(self, name):
        seq_dct = {}
        s = self.sequences[name]
        for method_name, v_names in s.items():
            val_list, label_lst = self.vectors.loop(v_names)
            val_list = [self.aliases.translate(d) for d in val_list]
            seq_dct[method_name] = {'val_list': val_list,
                                    'label_lst': label_lst}
        return seq_dct
    
    @property
    def sequence_names(self):
        return list(self.sequences.keys())
        
    def all_sequences(self):
        out = {}
        for seq_name in self.sequence_names:
            seq_dct = self.sequence(seq_name)
            out[seq_name] = seq_dct
        return out
        

